#!BPY
import bpy, bmesh
import mathutils
import math
import os
from collections import defaultdict
from mathutils import Vector
import addon_utils

###===================================================================================================
### CREATE FILEPATH
###===================================================================================================
dir = os.path.dirname( bpy.data.filepath )
name = bpy.path.basename( bpy.context.blend_data.filepath )
name = name.split( ".blend" )[0]
filepath = os.path.join( dir, name + ".h" )

###===================================================================================================
### GLOBALS
###===================================================================================================
tex_id = -1
img = []
img_id = 0
sprite_image = bpy.data.images.new( name="_tga", width=32, height=32, alpha=True )
    
###===================================================================================================
### TEXNO DEFINITION
###===================================================================================================
#texdef = str().join( name.split() ).upper() + "_TEXNO"

###===================================================================================================
### POPULATE WITH MODEL DATA
###===================================================================================================
def Export_Mesh( _tex_id ):
    ngons = False
    with open (filepath, 'w') as file:
        file.write( "/* Model Name: " + name.title() + " */\n" )
        file.write( "/* Total Objects : " + str( len( bpy.data.objects ) ) + " */\n" )
        file.write( "/*\n" )
        for ob in bpy.data.objects:
            file.write("    -" + ob.name + "\n")
        file.write( "*/\n" )
        file.write( "\n" )
        
        file.write( "#ifndef __3D_%s_H__\n" % ( name.upper() ) )
        file.write( "#define __3D_%s_H__\n" % ( name.upper() ) )
        file.write( "\n" )
        
        ###===========================================================================================
        ### GET DATA FROM EACH OBJECT IN THE SCENE
        ###===========================================================================================
        for ob in bpy.data.objects:
            ###=======================================================================================
            if ob.type == "MESH":
                file.write( "// %s\n" % ( ob.name.upper() ) )
                ###===================================================================================
                ### VERTICES / POINTS
                ###===================================================================================
                file.write( "static POINT point_" + ob.name + "[%d] = {\n" % ( len( ob.data.vertices ) ) )
                for v in ob.data.vertices:
                    x, y, z = v.co * 500000 ## Convert Units to Sega Saturn 3D Space
                    file.write( "   {%9.6f, %9.6f, %9.6f},\n" % ( x, y, z ) )
                file.write( "};\n\n" )
                
                ###===================================================================================
                ### FACES AND NORMALS / POLYGONS
                ###===================================================================================
                file.write( "static POLYGON polygon_" + ob.name + "[%d] = {\n" % ( len( ob.data.polygons ) ) )
                for p in ob.data.polygons:
                    if len( p.vertices ) == 4:
                        file.write( "   {{%9.6f, %9.6f, %9.6f}, {%3d, %3d, %3d, %3d}},\n" % (
                            p.normal.x,
                            p.normal.y,
                            p.normal.z,
                            p.vertices[3],
                            p.vertices[0],
                            p.vertices[1],
                            p.vertices[2]
                        ) )
                    elif len( p.vertices ) == 3:
                        file.write( "   {{%9.6f, %9.6f, %9.6f}, {%3d, %3d, %3d, %3d}},\n" % (
                            p.normal.x,
                            p.normal.y,
                            p.normal.z,
                            p.vertices[0],
                            p.vertices[1],
                            p.vertices[2],
                            p.vertices[0]
                        ) )
                    else:
                        file.write( "// CANNOT CONVERT THIS FACE!\n" )
                        ngons = True
                        p.select = True
                        
                file.write( "};\n\n" )
                
                ###===================================================================================
                ### ATTRIBUTES
                ###===================================================================================
                already_used_image_names = []
                file.write( "static ATTR attribute_" + ob.name + "[%d] = {\n" % ( len( ob.data.polygons ) ) )
                for p in ob.data.polygons:
                    ###===============================================================================
                    ### FACE COLORS
                    ###===============================================================================
                    
                    r = g = b = 31
                    
                    try:
                        material_for_poly = ob.data.materials[p.material_index] if p.material_index < len(ob.data.materials) else { "node_tree": False }
                        if material_for_poly.node_tree:
                            material_for_poly.use_nodes = True
                            bsdf = material_for_poly.node_tree.nodes["Principled BSDF"]
                            color_values = bsdf.inputs[0].default_value
                            r,g,b = ((color_values[0]*100) / 100) * 255, ((color_values[1]*100) / 100) * 255, ((color_values[2]*100) / 100) * 255
                    except IndexError:
                        pass
                            
                    ###================================================================================
                    
                    if any((node.type == 'TEX_IMAGE' and node.image != None) for node in ob.data.materials[p.material_index].node_tree.nodes):
                        
                        _tex_id += 1
                         
                        texno = False
                        colno = "CL32KRGB|No_Gouraud"
                        spr = "sprNoflip"
                    else:
                        texno = True
                        colno = "CL32KRGB|MESHoff"
                        spr = "sprPolygon"
                        
                    file.write( "   ATTRIBUTE(Dual_Plane, SORT_CEN, %20s, C_RGB(%2d, %2d, %2d), CL32KRGB | No_Gouraud, %27s, %10s, UseLight),\n" % (
                        _tex_id if texno == False else "No_Texture",
                        r,
                        g,
                        b,
                        colno,
                        spr
                        ) )
                file.write( "};\n\n" )
                    
                ###===================================================================================
                ### MESH STRUCTURE
                ###===================================================================================
                file.write( "jo_3d_mesh     mesh_" + ob.name + " = { \n")
                file.write( "   .data =\n" )
                file.write( "   {\n")
                file.write( "       point_" + ob.name + ",\n" )
                file.write( "       %3d,\n" % ( len( ob.data.vertices ) ) )
                file.write( "       polygon_" + ob.name + ",\n" )
                file.write( "       %3d,\n" % ( len( ob.data.polygons ) ) )
                file.write( "       attribute_" + ob.name + ",\n" )
                file.write( "   }\n" )
                file.write( "};\n\n" )
                
                ###===================================================================================
                ### MESH DRAW FUNCTION
                ###===================================================================================
                file.write( "static __jo_force_inline void      display_%s_mesh(void)" % ( ob.name ) )
                file.write( "{\n" )
                file.write( "   jo_3d_mesh_draw(&mesh_%s);\n" % ( ob.name ) )
                file.write( "}\n\n" )
        
        file.write( "#endif\n" )
        file.close()
        ###===========================================================================================

    if ngons == True:
        raise ValueError( "This model contains ngons. They have been automatically selected for review." )
    ###===============================================================================================

###===================================================================================================
### SPRITE TEXTURE CONVERSION
###===================================================================================================
def Export_Textures( _dir, _img_id, _sprite_image ):
    _dir += os.path.dirname( "/TEXTURES/" )
    if os.path.exists( _dir ) == False:
        os.mkdir( _dir )
    
    # Convert each textured plane as a TGA image
    bpy.context.scene.render.image_settings.file_format = "TARGA_RAW"
    _sprite_image.file_format = "TARGA"

    already_used_material_names = []
    for ob in bpy.data.objects:
        if ob.type == "MESH" and ob.data.uv_layers.active != None:
            
            bpy.ops.object.select_all( action="DESELECT" )
            ob.select_set(True)
            
            bpy.types.RenderSettings.bake_type = 'NORMALS'
#            bpy.types.RenderSettings.use_single_layer = True

            for p in ob.data.polygons:
                material_for_poly = { "node_tree": { "nodes": [] }, "name": "" }
                try:
                    material_for_poly = ob.data.materials[p.material_index] if p.material_index < len(ob.data.materials) else material_for_poly
                except IndexError:
                    pass
                
                if material_for_poly.name in already_used_material_names or "_" not in material_for_poly.name: continue
            
                uv_map_name = []
                if material_for_poly.node_tree:
                    # sort nodes based on bl_static_type and reverse the nodes so we get UVMAP before we get TEX_IMAGE
                    sorted_nodes = sorted(material_for_poly.node_tree.nodes, key=lambda n: n.bl_static_type, reverse=True)
                    for n in sorted_nodes:
                        if "UVMAP" in n.bl_static_type: uv_map_name.append(n.uv_map)
                        if "TEX_IMAGE" in n.bl_static_type and len(uv_map_name) > 0:
                            
                            current_uv_layer = ob.data.uv_layers[uv_map_name[0]]
                            current_uv_layer.active = True

                            texture_node = material_for_poly.node_tree.nodes.new('ShaderNodeTexImage')
                            texture_node.name = 'Bake_node'
                            texture_node.select = True
                            material_for_poly.node_tree.nodes.active = texture_node
                            texture_node.image = _sprite_image #Assign the image to the node
                            
                            current_uv_layer.active = True
                            current_uv_layer.active_clone = True
                            
                             # create new uv map/layer for baking do_init will copy the currently selected uvmap, so use False
                            new_uv_layer = ob.data.uv_layers.new(name="_uv_temp", do_init=False)
                            
                            possible_uvs = [current_uv_layer.data[l].uv for l in p.loop_indices]

                            # select new uv
                            new_uv_layer.active = True
                            new_uv_layer.active_render = True
       
                            # set uv map
                            ind = 0
                            for vert_idx, loop_idx in zip(p.vertices, p.loop_indices):
                                new_uv_layer.data[loop_idx].uv = possible_uvs[ind][:] # ie UV coord for each face with vert  me.vertices[vert_idx]
                                ind += 1

                            # Save our image as a new TGA file
                            # Format our file name to make it compatible with texture loading
                            img.append( _img_id )
                            filename = "%s.TGA" % ( _img_id )
                            _img_id += 1
                            filepath = os.path.join( _dir, filename )
                            
                            bpy.ops.object.bake(
                                                type="DIFFUSE", 
                                                pass_filter={"COLOR"}, 
                                                filepath=filepath, 
                                                width=32,
                                                height=32,
                                                target="IMAGE_TEXTURES", 
                                                save_mode="EXTERNAL",
                                                use_clear=True
                                            )
                            
                            _sprite_image.save_render(filepath=filepath)
                            
                            # Clear the image for the next one
                            bpy.data.images.remove( _sprite_image )
                            _sprite_image = bpy.data.images.new( name="_tga", width=32, height=32, alpha=True )
                            # In the last step, we are going to delete the nodes/uvmaps we created earlier
                            material_for_poly.node_tree.nodes.remove(texture_node)
                            ob.data.uv_layers.remove(new_uv_layer)
            
            
            bpy.ops.object.mode_set(mode = 'OBJECT')
            bpy.data.images.remove( _sprite_image )
            bpy.ops.object.select_all( action = "DESELECT" )
            ob.select_set(False)

###===================================================================================================
### TEXTURE HEADER
###===================================================================================================
def Export_TexHeader( _img_nums ):
    # Create a Header file that loads all of our textures
    filepath = os.path.join( dir, name + "_TEX.h" )

    with open ( filepath, 'w' ) as file :
        file.write( "#ifndef __%s_TEX__\n" % ( name ) )
        file.write( "#define __%s_TEX__\n" % ( name ) )
        file.write( "\n" )
        file.write( "void       Load_%s_Texture(void)\n" % ( name.replace(" ", "" ) ) )
        file.write( "{\n" )
        for i in _img_nums:
            file.write( "   jo_sprite_add_tga (JO_ROOT_DIR, \"%s.TGA\", JO_COLOR_Transparent);\n" % ( i ) )
        file.write( "}\n" )
        file.write( "#endif\n" )
        file.close()
        
###===================================================================================================
### ANIMATIONS
###===================================================================================================
def Export_Anims( ):
    return False

###===================================================================================================
### EXPORT SEQUENCE
###===================================================================================================
# Export Options
export_Mesh      = True
export_Textures  = True
export_TexHeader = True
export_Anims     = False

if export_Mesh:
    Export_Mesh( tex_id )
    Export_TexHeader( img )
    
if export_Textures:
    Export_Textures( dir, img_id, sprite_image )
    
if export_TexHeader and export_Textures:
    Export_TexHeader( img )

if export_Anims:
    Export_Anims( )

###===============================================================================================
### CLEAN UP
###===============================================================================================
#sprite_image.user_clear()
#bpy.data.images.remove( sprite_image )

###===============================================================================================